"""
Tketris

Tetris using tkinter

Author:  Anshul Kharbanda
Created: 10 - 11 - 2018
"""
from tkinter import *

class PlayingMenu(Frame):
    """
    The side menu shown during gameplay.
    """
    def __init__(self, master=None):
        """
        Initializes Frame
        """
        super(PlayingMenu, self).__init__(master)
        self.init_ui()

    def init_ui(self):
        """
        Initializes UI
        """
        self.score = Label(self, text="Score: 0")
        self.score.pack(fill=Y, expand=1)

        # Buttons frame
        self.buttons = Frame(self)
        self.pause = Button(self.buttons,
            text="Pause",
            command=self.toggle_game_resume)
        self.pause.pack()
        self.buttons.pack(fill=Y, expand=1)

    def display_score(self, score):
        """
        Display current score
        
        :param score: the current score to display
        """
        self.score.config(text="Score: {}".format(score))

    def display_game_resume_state(self, resume):
        """
        Display the appropriate pause/resume button text
        based on resume state
        
        :param resume: true if the game has resumed, false if paused
        """
        if resume:
            self.pause.config(text="Pause")
        else:
            self.pause.config(text="Resume")

    def toggle_game_resume(self):
        """
        Handle to toggle game pause or resume
        """
        self.master.toggle_game_resume()



"""
Tketris

Tetris using tkinter

Author:  Anshul Kharbanda
Created: 10 - 11 - 2018
"""
import numpy as np
from random import choice as random_choice
from ..numpy_algorithms import transform_tileset

"""
The tetris mino classes. In tketris, Minos are represented as objects 
containing the basic set of tiles comprising the mino shape, as well 
as the current position and orientation of the Mino. Mino objects 
have a method which automatically calculates the tiles to draw 
when rendering.
"""

class Mino:
    """
    Mino abstraction. This class contains the class method 'new_mino' which 
    generates a new subclass for a particular mino type given the type's 
    information, used below. This method also appends the new mino type to 
    the minos array, which is used in the 'random' method. 
    
    The 'random' method returns a random mino at the given position and 
    orientation (default at the top middle at orientation 0).
    """
    
    # The minos array. 'random' picks a mino type from this array
    minos = []

    @classmethod
    def new_type(cls, name, color, shape):
        """
        Creates a new mino type and saves it to the minos array.
        
        :param name: the name of the new mino type
        :param color: the tile color of the mino type
        :param shape: the shape tileset of the mino type (centered at 0,0)
        """
        mtype = type(name, (Mino,), {
            'color':color,
            'shape':np.array(shape, dtype=int)
        })
        cls.minos.append(mtype)
        return mtype

    @classmethod
    def random(cls, position=(0, 4), orientation=0):
        """
        Returns a random new mino at the given position and orientation
        
        :param position: initial position of new mino (default to top center)
        :param orientation: orientation of new mino (default to top center)
        
        :return: random new mino at position and orientation
        """
        return random_choice(cls.minos)(position, orientation)

    def __init__(self, position=(0, 4), orientation=0):
        """
        Initializes new Mino at given position and orientation
        
        :param position: initial position of new mino (default to top center)
        :param orientation: orientation of new mino (default to top center)
        """
        self.position = np.array(position)
        self.orientation = orientation

    @property
    def tiles(self):
        """
        The absolute tile positions of the mino
        """
        return transform_tileset(self.position, self.orientation, self.shape)

# Mino types
IMino = Mino.new_type('IMino', '#00bbff', [
    [-1, 0],
    [ 0, 0],
    [ 1, 0],
    [ 2, 0]
])
JMino = Mino.new_type('JMino', '#3300ff', [
    [-1,  0],
    [ 0,  0],
    [ 1,  0],
    [ 1, -1]
])
LMino = Mino.new_type('LMino', '#ff6600', [
    [-1,  0],
    [ 0,  0],
    [ 1,  0],
    [ 1,  1]
])
TMino = Mino.new_type('TMino', '#ff00ff', [
    [0, -1],
    [0,  0],
    [0,  1],
    [1,  0]
])
SMino = Mino.new_type('SMino', '#ff0000', [
    [1, -1],
    [1,  0],
    [0,  0],
    [0,  1]
])
ZMino = Mino.new_type('ZMino', '#00aa00', [
    [0, -1],
    [0,  0],
    [1,  0],
    [1,  1]
])
OMino = Mino.new_type('OMino', '#999900', [
    [0, -1],
    [0,  0],
    [1, -1],
    [1,  0]
])



"""
Tketris

Tetris using tkinter

Author:  Anshul Kharbanda
Created: 10 - 11 - 2018
"""
import numpy as np
from .bounds import TileSetBound

class BoardTileSet:
    """
    Represents all tiles currently on the board.
    Handles boundaries for tiles and row completion detection.
    """
    def __init__(self, tile_colors=[]):
        """
        Initializes instance with the given initial tile colors
        
        :param tile_colors: initial tile_colors
        """
        self.tile_colors = tile_colors

    def clear_tiles(self):
        """
        Clears all tiles from board
        """
        self.tile_colors = []

    def add_tiles(self, tiles, color):
        """
        Adds the given tiles in the given color to the tileset
        
        :param tiles: set of tiles to add
        :param color: color of new tiles being added
        """
        for tile in tiles.tolist():
            self.tile_colors.append((tile[0], tile[1], color))

    def clear_rows(self):
        """
        Clears all full rows.
        
        :return: number of rows cleared
        """
        cleared_rows = 0
        required_columns = {i for i in range(10)}
        rows = {tile[0] for tile in self.tile_colors}
        for row in rows:
            # Check if all required columns are in row
            columns = {tile[1] for tile in self.tile_colors if tile[0] == row}
            clear_row = all(required in columns for required in required_columns)
            if clear_row:
                self.tile_colors = [
                    (i + 1 if i < row else i, j, color)
                    for i, j, color in self.tile_colors if i != row
                ]
                cleared_rows += 1
        return cleared_rows

    @property
    def tiles(self):
        """
        All tiles in the tileset as a numpy tileset
        """
        return np.array([[i, j] for i, j, color in self.tile_colors])

    @property
    def left_bound(self):
        """
        Left boundary of tileset
        """
        return TileSetBound(self.tiles, 1, -1)

    @property
    def right_bound(self):
        """
        Right boundary of tileset
        """
        return TileSetBound(self.tiles, 1, 1)

    @property
    def up_bound(self):
        """
        Upper boundary of tileset
        """
        return TileSetBound(self.tiles, 0, -1)



"""
Tketris

Tetris using tkinter

Author:  Anshul Kharbanda
Created: 10 - 11 - 2018
"""
import numpy as np
from ..numpy_algorithms import tileset_intersection, tile_in_set

"""
This file defines the code which checks for boundaries through which the
current mino is unable to move.

This is implemented by computing all of the tiles that constitute a boundary in a
given direction. Since a mino can only move at right angles, if the mino occupies
any of these tiles, it is assumed the mino cannot move in that direction.

Before a mino is allowed to move, the control logic checks if the mino is occupying
the corresponding boundary. If so, it is not allowed to move in that direction
and will not respond to the command.
"""

class Bound:
    """
    Generic bound class
    """
    def collision(self, tiles):
        """
        Check if any of the given tiles is within any of the tiles in the bound

        :param tiles: the tiles being checked

        :return: True if any of the given tiles is within any of the tiles
                 in the bound
        """
        return tileset_intersection(tiles, self.tiles)

class TileSetBound(Bound):
    """
    Bound for a side of a set of tiles
    """
    def __init__(self, root_tiles, axis, direction):
        """
        Initializes instance
        """
        self.root_tiles = root_tiles
        self.axis = axis
        self.direction = direction

    @property
    def normal(self):
        """
        Docstring for direction_vector property
        """
        norm = np.array([ 0, 0 ])
        norm[self.axis] = self.direction
        return norm

    @property
    def tiles(self):
        """
        The tiles that occupy this bound
        """
        if self.root_tiles.shape[0] == 0:
            return np.array([])
        else:
            bounds = self.root_tiles + self.normal
            bound_is_open = np.logical_not(tile_in_set(bounds, self.root_tiles))
            return np.compress(bound_is_open, bounds, axis=0)

class BoardBound(Bound):
    """
    Game board bound
    """
    def __init__(self, axis, endpoint, size):
        """
        Initializes instance
        """
        self.axis = axis
        self.endpoint = endpoint
        self.size = size

    @property
    def tiles(self):
        """
        The tiles that occupy this bound
        """
        tiles = []
        for i in range(self.size):
            tile = [i, i]
            tile[self.axis] = self.endpoint
            tiles.append(tile)
        return np.array(tiles)

class BoardBounds:
    """
    All Tketris game board bounds
    """
    def __init__(self):
        """
        Initializes bounds
        """
        self.left_bound = BoardBound(1, 0, 20)
        self.right_bound = BoardBound(1, 9, 20)
        self.down_bound = BoardBound(0, 19, 10)



"""
Tketris

Tetris using tkinter

Author:  Anshul Kharbanda
Created: 10 - 11 - 2018
"""
from tkinter import *
import numpy as np

class Board(Frame):
    """
    Tketris Game Board where the mino magic happens!
    """
    bg_color = "white"
    clear_color = None
    box_size = 20

    def __init__(self, master=None):
        """
        Initializes Game Board
        """
        super(Board, self).__init__(master, bg=self.bg_color)
        self.init_ui()

    def init_ui(self):
        """
        Initializes User Interface
        """
        # Set color to background if color is not given
        if self.clear_color is None:
            self.clear_color = self.master['bg']

        # Set tiles
        self.tiles = []
        for i in range(20):
            row = []
            for j in range(10):
                tile = Canvas(self,
                    width=self.box_size,
                    height=self.box_size,
                    bg=self.clear_color)
                tile.grid(row=i, column=j)
                row.append(tile)
            self.tiles.append(row)

    def draw_tile(self, tile, color):
        """
        Draws the given tile
        """
        if 0 <= tile[0] and tile[0] <= 19 and 0 <= tile[1] and tile[1] <= 9:
            self.tiles[tile[0]][tile[1]].config(bg=color)

    def draw_tiles(self, tiles, color):
        """
        Draws all tiles
        """
        for tile in tiles.tolist():
            self.draw_tile(tile, color)

    def clear_tiles(self):
        """
        Sets all tiles to clear color
        """
        for row in self.tiles:
            for tile in row:
                tile.config(bg=self.clear_color)



"""
Tketris

Tetris using tkinter

Author:  Anshul Kharbanda
Created: 10 - 11 - 2018
"""
from tkinter import *

class GameOverMenu(Frame):
    """
    Side Menu shown on Game Over
    """
    def __init__(self, master=None):
        """
        Initializes Frame
        """
        super(GameOverMenu, self).__init__(master)
        self.init_ui()

    def init_ui(self):
        """
        Initializes UI
        """
        # Game over label
        self.game_over_label = Label(self, text="Game Over")
        self.game_over_label.pack(fill=Y, expand=1)
        self.final_score_label = Label(self, text="Final Score: 0")
        self.final_score_label.pack(fill=Y, expand=1)

        # Buttons frame
        self.buttons = Frame(self)
        self.reset_button = Button(self.buttons,
            text="Restart",
            command=self.reset_game)
        self.reset_button.pack()
        self.buttons.pack(fill=Y, expand=1)

    def display_final_score(self, final_score):
        """
        Display final score after game
        
        :param final_score: final score of the game
        """
        self.final_score_label.config(text="Final Score: {}".format(final_score))

    def reset_game(self):
        """
        Handle to start new game
        """
        self.master.reset_game()



"""
Tketris

Tetris using tkinter

Author:  Anshul Kharbanda
Created: 10 - 11 - 2018
"""
from tkinter import *
from .game_over import GameOverMenu
from .playing import PlayingMenu

class SideMenu(Frame):
    """
    Side Menu Frame which displays game state and score
    Handles actions like pause/resume and restart and
    switching between playing and game over menus
    """
    side_menu_color = "#ccc"
    current_menu = None

    def __init__(self, master=None):
        """
        Initializes Frame
        """
        super(SideMenu, self).__init__(master,
            bd=1,
            bg=self.side_menu_color,
            width=20)
        self.init_ui()

    def init_ui(self):
        """
        Initializes UI
        """
        self.select_playing()

    def select_menu(self, MenuType):
        """
        Select new menu to display
        """
        if self.current_menu:
            self.current_menu.destroy()
        self.current_menu = MenuType(self)
        self.current_menu.pack(fill=Y, expand=1, ipadx=10, ipady=10)

    def select_game_over(self):
        """
        Display game over menu
        """
        self.select_menu(GameOverMenu)

    def select_playing(self):
        """
        Display select playing menu
        """
        self.select_menu(PlayingMenu)

    def display_score(self, score):
        """
        Display current score
        
        :param score: current score of game
        """
        if type(self.current_menu) is PlayingMenu:
            self.current_menu.display_score(score)

    def display_game_resume_state(self, resume):
        """
        Display game resume state
        
        :param resume: true if game is resumed, false if paused
        """
        if type(self.current_menu) is PlayingMenu:
            self.current_menu.display_game_resume_state(resume)

    def display_final_score(self, final_score):
        """
        Display final game score
        
        :param final_score: final score of the game
        """
        if type(self.current_menu) is GameOverMenu:
            self.current_menu.display_final_score(final_score)

    def reset_game(self):
        """
        Handle for starting new game
        """
        self.master.reset_game()

    def toggle_game_resume(self):
        """
        Handle for toggle game resume
        """
        self.master.toggle_game_resume()



"""
Tketris

Tetris using tkinter

Author:  Anshul Kharbanda
Created: 10 - 11 - 2018
"""
import numpy as np

"""
These are the actions that the user can perform in the game They are grouped
into mixins by their required functionality. Each of these actions are bound
by a key. They have a condition function (preceding with "can"), and the
action is only performed if this condition is true.

Current actions are:
    - Rotate mino
    - Move mino left
    - Move mino right
    - Move mino down
"""

class Action:
    """
    MIXIN

    Generic action done on a key event
    """
    def can_do_action(self):
        """
        Docstring for can_do_action
        """
        return self.game_continue and self.game_resume

class RotateAction(Action):
    """
    MIXIN

    Rotate mino
    """
    def can_rotate(self):
        """
        Docstring for can_rotate
        """
        return self.can_do_action()

    def rotate(self, event):
        """
        Rotate current mino
        """
        if self.can_rotate():
            self.mino.orientation = (self.mino.orientation + 1) % 4
            self.render()

class MoveLeftAction(Action):
    """
    MIXIN

    Move mino left
    """
    def can_move_left(self):
        """
        True if the current mino can move left
        """
        return self.can_do_action() and not (
            self.bounds.left_bound.collision(self.mino.tiles) \
            or self.board_tileset.right_bound.collision(self.mino.tiles)
        )

    def move_left(self, event):
        """
        Move current mino left
        """
        if self.can_move_left():
            self.mino.position += np.array([0, -1])
            self.render()

class MoveRightAction(Action):
    """
    MIXIN

    Move mino right
    """
    def can_move_right(self):
        """
        True if the current mino can move right
        """
        return self.can_do_action() and not (
            self.bounds.right_bound.collision(self.mino.tiles) \
            or self.board_tileset.left_bound.collision(self.mino.tiles)
        )

    def move_right(self, event):
        """
        Move current mino right
        """
        if self.can_move_right():
            self.mino.position += np.array([0, 1])
            self.render()

class MoveDownAction(Action):
    """
    MIXIN

    Move mino down
    """
    def can_move_down(self):
        """
        True if the current mino can move down
        """
        return self.can_do_action() and not (
            self.bounds.down_bound.collision(self.mino.tiles) \
            or self.board_tileset.up_bound.collision(self.mino.tiles)
        )

    def move_down(self, event):
        """
        Move current mino down
        """
        if self.can_move_down():
            self.mino.position += np.array([1, 0])
            self.render()



"""
Tketris

Tetris using tkinter

Author:  Anshul Kharbanda
Created: 10 - 11 - 2018
"""
from . import Tketris
from tkinter import Tk
from argparse import ArgumentParser
from sys import argv

# Argument Parser
parser = ArgumentParser()
parser.add_argument('-d', '--debug', action='store_true')

# Main code
if __name__ == '__main__':
    args = parser.parse_args(argv[1:])
    root = Tk()
    app = Tketris(root)
    app.pack()
    app.DEBUG = args.debug
    root.title('Tketris')
    root.resizable(False, False)
    root.mainloop()



"""
Tketris

Tetris using tkinter

Author:  Anshul Kharbanda
Created: 10 - 11 - 2018
"""
import numpy as np

"""
Helper methods for tketris systems that use numpy (pretty much the tilesets).
"""

def tile_in_set(tiles_a, tiles_b):
    """
    Returns an array of booleans which indicate if the corresponding tile 
    in the first tileset is found in the second tileset

    :param tiles_a: the first tileset being checked
    :param tiles_b: the second tileset being checked

    :return: an array of booleans which are true if the tile in that row matches
             any of the tiles in the other tileset
    """
    # Return false if any of the arguments have no dimension
    if tiles_a.shape[0] == 0 or tiles_b.shape[0] == 0:
        return False
    else:
        # Expand arrays into third dimension
        expanded_a = np.expand_dims(tiles_a, axis=1)
        expanded_b = np.expand_dims(tiles_b, axis=0)

        # Check equality of point pairs
        point_pair_equal = np.all(np.equal(expanded_a, expanded_b), axis=2)

        # Return flattened array of checks for each point pair
        return np.any(point_pair_equal, axis=1).flatten()


def tileset_intersection(tiles_a, tiles_b):
    """
    Checks for an intersection in the given tiles, i.e. if any tile in the
    first tileset matches a tile in the second tileset

    :param tiles_a: the first tileset being checked
    :param tiles_b: the second tileset being checked

    :return: true if there is an intersection
    """
    # Check for tiles in set
    tiles_in_set = tile_in_set(tiles_a, tiles_b)

    # Check if any point pairs are equal
    return np.any(tiles_in_set)


def transform_tileset(pos, rot, tiles):
    """
    Returns the transformed tileset, at the correct position and rotation

    :param pos: the position offset of the tiles
    :param rot: the rotation integer of the tiles
    :param tiles: the tiles to transform

    :return: transformed tiles
    """
    # Transform matrix
    transform_matrix = np.matrix([
        [ np.cos(rot*np.pi/2), np.sin(rot*np.pi/2), 0],
        [-np.sin(rot*np.pi/2), np.cos(rot*np.pi/2), 0],
        [ pos[0],              pos[1],              1]
    ])

    # Homogenous tiles array
    one_column = np.ones((tiles.shape[0], 1))
    homogenous_tiles = np.concatenate((tiles, one_column), axis=1)

    # Multiply homogenous vector array by transform matrix
    transformed_homogenous_tiles = np.matmul(homogenous_tiles, transform_matrix)

    # Return tiles extracted from transformed homogenous tiles
    return transformed_homogenous_tiles[:,0:2].astype(int)



"""
Tketris

Tetris using tkinter

Author:  Anshul Kharbanda
Created: 10 - 11 - 2018
"""
import numpy as np
from .bounds import BoardBounds, TileSetBound
from .tileset import BoardTileSet
from .actions import RotateAction, MoveLeftAction, MoveRightAction, MoveDownAction
from .mino import Mino

class GameLogic(RotateAction, MoveLeftAction, MoveRightAction, MoveDownAction):
    """
    MIXIN

    Main Tketris Game Logic. All of the magic happens here!
    Well... most of the magic... just the game logic, as the name suggests...
    """
    def init_game(self):
        """
        Initialize game logic
        """
        self.bounds = BoardBounds()
        self.board_tileset = BoardTileSet([])
        self.score = 0

    def render(self):
        """
        Render UI
        """
        # Clear all tiles in board
        self.board.clear_tiles()

        # If debugging
        if self.DEBUG:
            # Draw boundaries of board
            self.board.draw_tiles(self.bounds.left_bound.tiles, '#faa')
            self.board.draw_tiles(self.bounds.right_bound.tiles, '#afa')
            self.board.draw_tiles(self.bounds.down_bound.tiles, '#aaf')

            # Draw boundaries of board tiles
            self.board.draw_tiles(self.board_tileset.left_bound.tiles, '#faa')
            self.board.draw_tiles(self.board_tileset.right_bound.tiles, '#afa')
            self.board.draw_tiles(self.board_tileset.up_bound.tiles, '#aaf')

        # Render tiles on ground
        for i, j, color in self.board_tileset.tile_colors:
            self.board.draw_tile((i, j), color)

        # Render current mino
        self.board.draw_tiles(self.mino.tiles, self.mino.color)

    def clear_tiles(self):
        """
        Clear tiles on the board
        """
        self.board_tileset.clear_tiles()

    def add_mino_to_board(self):
        """
        Adds mino to board tileset, clears required rows,
        and adds row score to the total score
        """
        self.board_tileset.add_tiles(self.mino.tiles, self.mino.color)
        rows = self.board_tileset.clear_rows()
        self.row_score(rows)

    def new_round(self):
        """
        Creates a new mino, starting a new round
        (trigger game over if new mino can't move down)
        """
        self.mino = Mino.random()
        self.render()
        if not self.can_move_down():
            self.game_continue = False
            self.on_game_over()

    def soft_drop(self):
        """
        Add soft drop score
        """
        self.score += 4
        self.on_update_score()

    def row_score(self, rows):
        """
        Compute row score given number of cleared rows

        :param rows: number of cleared rows
        """
        if rows == 1:
            self.score += 40
        elif rows == 2:
            self.score += 100
        elif rows == 3:
            self.score += 300
        elif rows == 4:
            self.score += 1200

        # Only call update score if score was actually updated
        if rows > 0:
            self.on_update_score()

    def clock_tick_update(self):
        """
        Update every clock tick. If mino can't move down, add soft drop score,
        add mino to board, and start new round. Else just move the mino down.
        """
        if not self.can_move_down():
            self.soft_drop()
            self.add_mino_to_board()
            self.new_round()
        else:
            self.move_down(None)

    def start_game(self):
        """
        Start new game. Reset board and score. Set flags. Start new round.
        """
        self.score = 0
        self.board_tileset.clear_tiles()
        self.game_continue = True
        self.game_resume = True
        self.new_round()
        self.on_start_game()

    def on_start_game(self):
        """
        HOOK

        Called after game starts
        """
        pass

    def on_game_over(self):
        """
        HOOK

        Called after game over
        """
        pass

    def on_update_score(self, score):
        """
        HOOK

        Called after score updates
        """
        pass



"""
Tketris

Tetris using tkinter

Author:  Anshul Kharbanda
Created: 10 - 11 - 2018
"""
from distutils.core import setup

# Setup function
setup(name='tketris',
      version='1.0',
      description='Tetris written in python and tkinter',
      author='Anshul Kharbanda',
      author_email='akanshul97@gmail.com',
      homepage='http://github.com/andydevs/tketris',
      packages=[
        'tketris',
        'tketris.game',
        'tketris.view',
        'tketris.view.side_menu'
      ],
      install_requires=['numpy'])



"""
Tketris

Tetris using tkinter

Author:  Anshul Kharbanda
Created: 10 - 11 - 2018
"""
from .game import GameLogic
from .view.board import Board
from .view.side_menu import SideMenu
from tkinter import *

class Tketris(Frame, GameLogic):
    """
    The main application frame. Includes the GameLogic mixin
    """
    DEBUG = False

    def __init__(self, master=None):
        """
        Initializes application instance
        """
        super(Tketris, self).__init__(master)
        self.init_ui()
        self.init_events()
        self.init_game()
        self.start_game()

    def init_ui(self):
        """
        Initializes user interface
        """
        self.board = Board(self)
        self.board.pack(side=LEFT, fill=Y)
        self.side_menu = SideMenu(self)
        self.side_menu.pack(side=RIGHT, fill=Y)

    def init_events(self):
        """
        Initializes keybound events
        """
        self.master.bind('<Up>', self.rotate)
        self.master.bind('<Left>', self.move_left)
        self.master.bind('<Right>', self.move_right)
        self.master.bind('<Down>', self.move_down)

    def reset_game(self):
        """
        Resets game
        """
        if not self.game_continue:
            self.start_game()

    def toggle_game_resume(self):
        """
        Toggle game resume state and update menu
        """
        self.game_resume = not self.game_resume
        self.side_menu.display_game_resume_state(self.game_resume)

    def run_clock(self):
        """
        Runs clock
        """
        if self.game_continue:
            if self.game_resume:
                self.clock_tick_update()
            self.master.after(1000, self.run_clock)

    def on_start_game(self):
        """
        Select playing and start clock on start game
        """
        self.side_menu.select_playing()
        self.run_clock()

    def on_game_over(self):
        """
        Displays game over menu and display final score on game over
        """
        self.side_menu.select_game_over()
        self.side_menu.display_final_score(self.score)

    def on_update_score(self):
        """
        Display score on playing menu on update score
        """
        self.side_menu.display_score(self.score)
